﻿window.sf = window.Syncfusion = window.Syncfusion || {};

(function ($, sf, undefined) {
    'use strict';

    sf.DataManagar = function (dataSource, query) {
        if (!(this instanceof sf.DataManagar))
            return new sf.DataManagar(dataSource, query);

        if (!dataSource)
            throw "DataManger : a dataSource is required to create Manager";

        var data = [];

        if (dataSource instanceof Array) {
            // JSON array
            data = {
                json: dataSource,
                offline: true
            };

        } else if (typeof dataSource === "object") {
            if (dataSource.url || dataSource.json) {
                data = {
                    url: dataSource.url,
                    json: dataSource.json,
                    headers: dataSource.headers,
                    accept: dataSource.accept,
                    data: dataSource.data,
                    requestType: dataSource.requestType,
                    key: dataSource.key,
                    crossDomain: dataSource.crossDomain || false,
                    jsonp: dataSource.jsonp || "$callback",
                    dataType: dataSource.dataType,
                    offline: dataSource.offline !== undefined ? dataSource.offline : dataSource.url ? false : true,
                    requiresFormat: dataSource.requiresFormat
                };
            }
        } else if (typeof dataSource === "string") {
            data = {
                url: dataSource,
                crossDomain: false,
                offline: false
            };
        }

        if (data.accept === undefined)
            data.accept = oData.accept;

        if (data.requiresFormat === undefined)
            data.requiresFormat = data.crossDomain;

        if (data.dataType === undefined)
            data.dataType = data.crossDomain ? "jsonp" : "json";

        this.dataSource = data;
        this.defaultQuery = query;
        this.adaptor = data.offline ? new sf.JsonAdaptor() : new sf.ODataAdaptor();
        return this;
    };

    sf.DataManagar.prototype = {
        setDefaultQuery: function (query) {
            this.defaultQuery = query;
        },

        executeQuery: function (query, done, fail, always) {
            if (typeof query === "function") {
                always = fail;
                fail = done;
                done = query;
                query = null;
            }

            if (!query)
                query = this.defaultQuery;

            if (!(query instanceof sf.Query))
                throw "DataManager - executeQuery() : A query is required to execute";

            var deffered = $.Deferred();

            deffered.then(done, fail, always);
            var args = { query: query };

            if (!this.dataSource.offline) {
                this._makeRequest(this.adaptor.getUrl(query, this.dataSource), deffered, args, query);
            } else {
                nextTick(function () {
                    var res = this.executeLocal(query);
                    args.result = res.result;
                    args.count = res.count;
                    args.getTableModel = getTableModel(query._fromTable, args.result, this);

                    deffered.resolveWith(this, [args]);
                }, this);
            }

            return deffered.promise();
        },

        executeLocal: function (query) {
            if (!this.defaultQuery && !(query instanceof sf.Query))
                throw "DataManager - executeLocal() : A query is required to execute";

            if (!this.dataSource.json)
                throw "DataManager - executeLocal() : Json data is required to execute";

            var ret, key;
            query = query || this.defaultQuery;

            // this does not support for URL binding
            // this simple array copy not object copy
            var result = this.dataSource.json.slice(0), count = result.length, cntFlg = true;

            for (var i = 0; i < query.queries.length; i++) {
                key = query.queries[i];
                ret = this.adaptor[key.fn].call(this, result, query, key.e);
                result = ret !== undefined ? ret : result;

                if (key.fn === "onPage" || key.fn === "onSkip" || key.fn === "onTake" || key.fn === "onRange") cntFlg = false;

                if (cntFlg) count = result.length;
            }

            if (query._subQuery) {
                var from = query._subQuery._fromTable, lookup = query._subQuery._lookup;

                if (lookup && lookup instanceof Array) {
                    buildHierarchy(query._subQuery._fKey, from, result, lookup);
                }

                for (var j = 0; j < result.length; j++)
                    result[i][from] = query._subQuery.using(sf.DataManagar(result[i][from])).executeLocal();
            }
            if (query._requiresCount) {
                return {
                    result: result,
                    count: count
                };
            } else
                return result;
        },

        _makeRequest: function (url, deffered, args, query) {
            var isSelector = !!query._subQuerySelector;

            var fnFail = $.proxy(function (e) {
                args.error = e;
                deffered.rejectWith(this, [args]);
            }, this);

            var process = $.proxy(function (data, count) {

                if (query._groupedColumns.length) {
                    for (var i = 0; i < query._groupedColumns.length; i++)
                        data = sf.group(data, query._groupedColumns[i]);
                }

                if (isSelector) return;

                args.count = parseInt(count, 10);
                args.result = data;
                args.getTableModel = getTableModel(query._fromTable, data, this);
                deffered.resolveWith(this, [args]);

            }, this);

            var fnQueryChild = $.proxy(function (data, selector) {
                var subDeffer = $.Deferred(),
                    childArgs = { parent: args };

                query._subQuery._isChild = true;

                var subUrl = this.adaptor.getUrl(query._subQuery, this.dataSource, data || selector);

                var childReq = this._makeRequest(subUrl, subDeffer, childArgs, query._subQuery);

                // TODO: Need to have fail handler

                subDeffer.done(function (subData) {
                    if (data) {
                        buildHierarchy(query._subQuery._fKey, query._subQuery._fromTable, data, subData);
                        process(data);
                    }
                });

                return childReq;
            }, this);
            var fnSuccess = $.proxy(function (data) {
                var count = 0;
                data = p.parseJson(data);

                if (data.__count) count = data.__count;
                if (data.d) data = data.d;
                if (data.__count) count = data.__count;

                if (data.results) data = data.results;

                if (!query._subQuery) {
                    process(data, count);
                    return;
                }

                if (!isSelector)
                    fnQueryChild(data);

            }, this);

            var req = $.ajax({
                url: url,
                dataType: this.dataSource.dataType,
                jsonp: this.dataSource.jsonp,
                cache: true,
                beforeSend: $.proxy(function (request) {
                    request.setRequestHeader("MaxDataServiceVersion", "2.0");

                    var headers = this.dataSource.headers;
                    for (var i = 0; headers && i < headers.length; i++) {
                        request.setRequestHeader(headers[i].name, headers[i].value);
                    }
                }, this),
                processData: false,
                success: fnSuccess,
                error: fnFail
            });

            if (isSelector) {
                var res = query._subQuerySelector.call(this, { query: query._subQuery, parent: query });

                if (res && res.length) {
                    req = $.when(req, fnQueryChild(null, res));

                    req.then(function (pData, cData) {
                        console.log("when");

                        pData = getResult(pData[0]);
                        cData = getResult(cData[0]);

                        buildHierarchy(query._subQuery._fKey, query._subQuery._fromTable, pData, cData);
                        isSelector = false;
                        process(pData);

                    }, function () {

                    });
                } else {
                    isSelector = false;
                }
            }

            return req;
        },

        saveChanges: function (changes, key, tableName) {
            var url = this.dataSource.url.replace(/\/*$/, '/' + oData.batch);
            var args = {
                url: tableName,
                key: key || this.dataSource.key
            };

            var req = this.adaptor.generateBatchRequest(changes, args);
            var deff = $.Deferred();
            $.ajax({
                type: "POST",
                url: url,
                beforeSend: $.proxy(function (request) {
                    request.setRequestHeader("Accept", oData.multipartAccept);
                    request.setRequestHeader("MaxDataServiceVersion", "2.0");
                    request.setRequestHeader("DataServiceVersion", "1.0");

                    var headers = this.dataSource.headers;
                    for (var i = 0; headers && i < headers.length; i++) {
                        request.setRequestHeader(headers[i].name, headers[i].value);
                    }
                }, this),
                contentType: "multipart/mixed; charset=UTF-8;boundary=" + args.guid,
                data: req,
                success: function (data, status, xhr) {
                    var guid = xhr.getResponseHeader("Content-Type"), cIdx, jsonObj;
                    guid = guid.substring(guid.indexOf("=batchresponse") + 1);
                    data = data.split(guid);
                    if (data.length < 2) return;

                    data = data[1];
                    var exVal = /(?:\bContent-Type.+boundary=)(changesetresponse.+)/i.exec(data);
                    data.replace(exVal[0], "");
                    
                    var changeGuid =exVal[1];
                    data = data.split(changeGuid);
                    
                    for (var i = data.length; i > -1; i--) {
                        if (!/\bContent-ID:/i.test(data[i]) || !/\bHTTP.+201/.test(data[i]))
                            continue;

                        cIdx = parseInt(/\bContent-ID: (\d+)/i.exec(data[i])[1]);
                        
                        if (changes.added[cIdx]) {
                            jsonObj = p.parseJson(/^\{.+\}/m.exec(data[i])[0]);
                            $.extend(changes.added[cIdx], getResult(jsonObj));
                        }
                    }

                    deff.resolveWith(this);
                },
                fail: function (e) {
                    deff.rejectWith(this, e);
                }
            });

            return deff;
        },

        insert: function (data, tableName) {
            if (this.dataSource.offline) {
                return this.dataSource.push(data);
            }

            var deffer = $.Deferred();

            $.ajax({
                type: "POST",
                contentType: "application/json; charset=utf-8",
                url: this.dataSource.url.replace(/\/*$/, tableName ? '/' + tableName : ''),
                data: JSON.stringify(data),
                dataType: "json",
                processData: false,
                beforeSend: $.proxy(function (request) {
                    request.setRequestHeader("MaxDataServiceVersion", "2.0");

                    var headers = this.dataSource.headers;
                    for (var i = 0; headers && i < headers.length; i++) {
                        request.setRequestHeader(headers[i].name, headers[i].value);
                    }
                }, this),
                success: function (record) {
                    record = getResult(p.parseJson(record));
                    deffer.resolveWith(this, [{ record: record, dataManager: this }]);
                },
                error: function (e) {
                    deffer.rejectWith(this, [{ error: error, dataManager: this }]);
                }
            });

            return deffer.promise();
        },

        remove: function (keyField, value, tableName) {
            if (this.dataSource.offline) {
                var ds = this.dataSource.json, i, res = null;

                if (typeof value === "object")
                    value = value[keyField];

                for (i = 0; i < ds.length; i++) {
                    if (ds[i][keyField] === value) break;
                }

                if (i !== ds.length) {
                    res = ds.splice(i, 1);
                }

                return res;
            }

            var url = this.dataSource.url.replace(/\/*$/, tableName ? '/' + tableName : '') + '(' + value + ')';

            return $.ajax({
                type: "DELETE",
                contentType: "application/json; charset=utf-8",
                url: url,
                beforeSend: $.proxy(function (request) {
                    request.setRequestHeader("MaxDataServiceVersion", "2.0");

                    var headers = this.dataSource.headers;
                    for (var i = 0; headers && i < headers.length; i++) {
                        request.setRequestHeader(headers[i].name, headers[i].value);
                    }
                }, this),
                dataType: "json"
            });
        },

        update: function (keyField, value, tableName) {
            var key = value;
            if (typeof value === "object")
                key = value[keyField];

            if (this.dataSource.offline) {
                var ds = this.dataSource.json, i, res = null;

                for (i = 0; i < ds.length; i++) {
                    if (ds[i][keyField] === key) break;
                }

                if (i !== ds.length) {
                    for (var prop in value) {
                        ds[i][prop] = value[prop];
                    }
                    res = ds[i];
                }

                return res;
            }
            var url = this.dataSource.url.replace(/\/*$/, tableName ? '/' + tableName : '') + '(' + key + ')';

            return $.ajax({
                type: "PUT",
                contentType: "application/json; charset=utf-8",
                url: url,
                dataType: "json",
                beforeSend: $.proxy(function (request) {
                    request.setRequestHeader("MaxDataServiceVersion", "2.0");

                    var headers = this.dataSource.headers;
                    for (var i = 0; headers && i < headers.length; i++) {
                        request.setRequestHeader(headers[i].name, headers[i].value);
                    }
                }, this),
                data: JSON.stringify(value),
                accept: oData.accept
            });
        }
    };

    var getResult = function (data) {
        if (data.d) data = data.d;
        if (data.results) data = data.results;
        return data;
    };

    var buildHierarchy = function (fKey, from, source, lookup) {
        var i, grp = {}, t;
        if (lookup.result) lookup = lookup.result;

        if (lookup.GROUPGUID) {
            for (i = 0; i < lookup.length; i++) {
                grp[lookup[i].key] = lookup[i].items;
                lookup[i].items.level -= 1;
            }
        } else {
            for (i = 0; i < lookup.length; i++) {
                t = grp[lookup[i][fKey]] || (grp[lookup[i][fKey]] = []);

                t.push(lookup[i]);
            }
        }

        for (i = 0; i < source.length; i++) {
            source[i][from] = grp[source[i][fKey]];
        }
    };

    var oData = {
        accept: "application/json;odata=light;q=1,application/json;odata=verbose;q=0.5",
        multipartAccept: "multipart/mixed",
        sortBy: "$orderby=",
        select: "$select=",
        skip: "$skip=",
        take: "$top=",
        count: "$inlinecount=",
        where: "$filter=",
        expand: "$expand=",
        batch: "$batch",
        changeSet: "--changeset_",
        batchPre: "batch_",
        contentId: "Content-Id: ",
        batchContent: "Content-Type: multipart/mixed; boundary=",
        changeSetContent: "Content-Type: application/http\nContent-Transfer-Encoding: binary ",
        batchChangeSetContentType: "Content-Type: application/json; charset=utf-8 "
    };

    var p = {
        parseJson: function (jsonText) {
            if (typeof jsonText === "string") {
                jsonText = JSON.parse(jsonText, p.jsonReviver);
            } else if (jsonText instanceof Array) {
                p.iterateAndReviveArray(jsonText);
            } else
                p.iterateAndReviveJson(jsonText);
            return jsonText;
        },
        iterateAndReviveArray: function (array) {
            for (var i = 0; i < array.length; i++)
                p.iterateAndReviveJson(array[i]);
        },
        iterateAndReviveJson: function (json) {
            var value;

            for (var prop in json) {
                if (prop.startsWith("__"))
                    continue;

                value = json[prop];
                if (typeof value === "object") {
                    if (value instanceof Array)
                        p.iterateAndReviveArray(value);
                    else
                        p.iterateAndReviveJson(value);
                } else
                    json[prop] = p.jsonReviver(prop, value);
            }
        },
        jsonReviver: function (field, value) {
            if (typeof value === "string") {
                var ms = /^\/Date\(([0-9]+)([+-][0-9]{4})?\)\/$/.exec(value);
                if (ms)
                    return new Date(parseInt(ms[1]));
            }
            return value;
        }
    };

    sf.Query = function (from) {
        if (!(this instanceof sf.Query))
            return new sf.Query(from);

        this.queries = [];
        this._key = "";
        this._fKey = "";

        if (typeof from === "string")
            this._fromTable = from || "";
        else if (from && from instanceof Array) {
            this._lookup = from;
        }
        this._expands = [];
        this._sortedColumns = [];
        this._groupedColumns = [];
        this._filters = [];
        this._sortKey = {};
        this._filterKey = {};
        this._subQuery = null;
        this._isChild = false;
        return this;
    };

    sf.Query.prototype = {
        key: function (field) {
            if (typeof field === "string")
                this._key = field;

            return this;
        },

        using: function (dataManager) {
            if (dataManager instanceof sf.DataManagar) {
                this.dataManagar = dataManager;
                return this;
            }

            throw "Query - using() : 'using' function should be called with parameter of instance sf.DataManager";
        },

        execute: function (dataManager, done, fail, always) {
            dataManager = dataManager || this.dataManagar;

            if (dataManager && dataManager instanceof sf.DataManagar)
                return dataManager.executeQuery(this, done, fail, always);

            throw "Query - execute() : dataManager needs to be is set using 'using' function or should be passed as argument";
        },

        executeLocal: function (dataManager) {
            // this does not support for URL binding

            dataManager = dataManager || this.dataManagar;

            if (dataManager && dataManager instanceof sf.DataManagar)
                return dataManager.executeLocal(this);

            throw "Query - executeLocal() : dataManager needs to be is set using 'using' function or should be passed as argument";
        },

        from: function (tableName) {
            if (typeof tableName === "string")
                this._fromTable = tableName;

            return this;
        },

        expand: function (tables) {
            if (typeof tables === "string")
                this._expands = [].slice.call(arguments, 0);
            else
                this._expands = tables.slice(0);

            return this;
        },

        where: function (fieldName, operator, value, ignoreCase) {
            operator = (operator || sf.FilterOperators.equal).toLowerCase();
            var comparer = null, isPredicate;

            if (typeof fieldName === "string")
                comparer = sf.data.fnOperators[operator] || sf.data.fnOperators.processSymbols(operator);
            else if (fieldName instanceof sf.Predicate) {
                isPredicate = true;
            } else {
                throw "Query - where : Invalid arguments";
            }

            if (isPredicate) {
                this._filters.push(fieldName);
            } else {
                this._filters.push({ field: fieldName, operator: operator, value: value, ignoreCase: ignoreCase });
            }
            this.queries.push({
                fn: "onWhere",
                e: {
                    isPredicate: isPredicate,
                    fieldName: fieldName,
                    operator: operator,
                    value: value,
                    ignoreCase: ignoreCase,
                    comparer: comparer
                }
            });
            return this;
        },

        search: function (searchKey, fieldNames, operator, ignoreCase) {
            if (!fieldNames || typeof fieldNames === "boolean") {
                fieldNames = [];
                ignoreCase = fieldNames;
            } else if (typeof fieldNames === "string")
                fieldNames = [fieldNames];

            if (typeof operator === "boolean") {
                ignoreCase = operator;
                operator = null;
            }
            operator = operator || sf.FilterOperators.contains;
            var comparer = sf.data.fnOperators[operator] || sf.data.fnOperators.processSymbols(operator);

            if (fieldNames.length) {
                var filter = [];
                for (var i = 0; i < fieldNames.length; i++) {
                    filter.push({ field: fieldNames[i], operator: operator, value: searchKey, ignoreCase: ignoreCase });
                }
                this._filters.push(filter);
            }
            this.queries.push({
                fn: "onSearch",
                e: {
                    fieldNames: fieldNames,
                    operator: operator,
                    searchKey: searchKey,
                    ignoreCase: ignoreCase,
                    comparer: comparer
                }
            });
            return this;
        },

        sortBy: function (fieldName, comparer, isFromGroup) {
            var order = sf.sortOrder.Ascending, isArray = fieldName instanceof Array, isFunc = false;
            if (typeof fieldName === "function")
                isFunc = true;

            if (typeof comparer === "boolean")
                comparer = !comparer ? sf.sortOrder.Ascending : sf.sortOrder.Descending;

            if (!comparer || typeof comparer === "string") {
                order = comparer ? comparer.toLowerCase() : sf.sortOrder.Ascending;
                comparer = sf.pvt.fnSort(comparer);
            }

            if (fieldName && !isFunc) {
                pushOdSortColumns(this._sortedColumns, fieldName, order, isFromGroup);

                if (!isArray)
                    fieldName = [fieldName];
            }

            this.queries.push({
                fn: "onSortBy",
                e: {
                    fieldName: fieldName,
                    comparer: comparer
                }
            });

            return this;
        },

        sortByDesc: function (fieldName) {
            return this.sortBy(fieldName, sf.sortOrder.Descending);
        },

        group: function (fieldName) {
            if (fieldName) {
                this._groupedColumns.push(fieldName);

                pushOdSortColumns(this._sortedColumns, this._groupedColumns.slice().reverse(), sf.sortOrder.Ascending, true);
            }
            this.queries.push({
                fn: "onGroup",
                e: {
                    fieldName: fieldName
                }
            });
            return this;
        },

        page: function (pageIndex, pageSize) {
            this.queries.push({
                fn: "onPage",
                e: {
                    pageIndex: pageIndex,
                    pageSize: pageSize
                }
            });
            return this;
        },

        range: function (start, end) {
            if (typeof start !== "number" || typeof end !== "number")
                throw "Query() - range : Arguments type should be a number";

            this.queries.push({
                fn: "onRange",
                e: {
                    start: start,
                    end: end
                }
            });
            return this;
        },

        take: function (nos) {
            if (typeof nos !== "number")
                throw "Query() - Take : Argument type should be a number";

            this.queries.push({
                fn: "onTake",
                e: {
                    nos: nos
                }
            });
            return this;
        },

        skip: function (nos) {
            if (typeof nos !== "number")
                throw "Query() - Skip : Argument type should be a number";

            this.queries.push({
                fn: "onSkip",
                e: { nos: nos }
            });
            return this;
        },

        select: function (fieldNames) {
            if (typeof fieldNames === "string")
                fieldNames = [].slice.call(arguments, 0);

            if (!(fieldNames instanceof Array)) {
                throw "Query() - Select : Argument type should be String or Array";
            }

            this.queries.push({
                fn: "onSelect",
                e: { fieldNames: fieldNames }
            });
            return this;
        },

        hierarchy: function (query, selectorFn) {
            if (!query || !(query instanceof sf.Query))
                throw "Query() - hierarchy : query must be instance of sf.Query";

            if (typeof selectorFn === "function")
                this._subQuerySelector = selectorFn;

            this._subQuery = query;
            return this;
        },

        foreignKey: function (key) {
            if (typeof key === "string")
                this._fKey = key;

            return this;
        },

        requiresCount: function () {
            this._requiresCount = true;

            return this;
        }
    };

    sf.ODataAdaptor = function () {
    };

    sf.ODataAdaptor.prototype = {
        onWhere: function (ds, self) {
            return self._filterKey;
        },
        onSearch: function (ds, self, e) {
            if (e.fieldNames.length === 0)
                throw "Query() - Search : oData search requires list of field names to search";

            return self._filterKey;
        },
        onSortBy: function (ds, self) {
            return self._sortKey;
        },
        onGroup: function (ds, self) {
            return self._sortKey;
        },
        onPage: function (ds, self, e) {
            var start = (e.pageIndex - 1) * e.pageSize, page = oData.skip;

            page += start + "&";
            page += oData.take;
            return page + e.pageSize;
        },
        onRange: function (ds, self, e) {
            var page = oData.skip;
            page += e.start + "&";
            page += e.oData.take;
            return page + (e.start - e.end);
        },
        onTake: function (ds, self, e) {
            return oData.take + e.nos;
        },
        onSkip: function (ds, self, e) {
            return oData.skip + e.nos;
        },
        onSelect: function (ds, self, e) {
            var select = oData.select;
            for (var i = 0; i < e.fieldNames.length; i++) {
                select += toODataProperty(e.fieldNames[i]);
                if (i != e.fieldNames.length - 1)
                    select += ",";
            }

            return select;
        },
        getUrl: function (query, ds, filters) {

            var q = queryGen(query, this).addFilters(filters);

            return q.buildUrl(ds.url, ds.requiresFormat);
        },
        generateBatchRequest: function (changes, e) {
            var initialGuid = e.guid = sf.getGuid(oData.batchPre);
            var args = {
                url: e.url,
                key: e.key,
                cid: 1,
                cSet: sf.getGuid(oData.changeSet)
            };
            var req = "--" + initialGuid + "\n";

            req += "Content-Type: multipart/mixed; boundary=" + args.cSet.replace("--", "") + "\n";

            req += this.generateInsertRequest(changes.added, args);
            req += this.generateUpdateRequest(changes.changed, args);
            req += this.generateDeleteRequest(changes.deleted, args);

            req += args.cSet + "--\n";

            return req + "--" + initialGuid + "--";
        },
        generateDeleteRequest: function (arr, e) {
            if (!arr) return "";
            var req = "";

            for (var i = 0; i < arr.length; i++) {
                req += "\n" + e.cSet + "\n";
                req += oData.changeSetContent + "\n\n";
                req += "DELETE ";
                req += e.url + "(" + arr[i].get(e.key) + ") HTTP/1.1\n";
                req += "Accept: " + oData.accept + "\n";
                req += oData.batchChangeSetContentType + "\n";
            }

            return req + "\n";
        },
        generateInsertRequest: function (arr, e) {
            if (!arr) return "";
            var req = "";

            for (var i = 0; i < arr.length; i++) {
                req += "\n" + e.cSet + "\n";
                req += oData.changeSetContent + "\n\n";
                req += "POST ";
                req += e.url + " HTTP/1.1\n";
                req += "Accept: " + oData.accept + "\n";
                req += "Content-Id: " + i + "\n";
                req += oData.batchChangeSetContentType + "\n\n";

                req += JSON.stringify(arr[i]) + "\n";
            }

            return req;
        },
        generateUpdateRequest: function (arr, e) {
            if (!arr) return "";
            var req = "";

            for (var i = 0; i < arr.length; i++) {
                req += "\n" + e.cSet + "\n";
                req += oData.changeSetContent + "\n\n";
                req += "PUT ";
                req += e.url + "(" + arr[i][e.key] + ")" + " HTTP/1.1\n";
                req += "Accept: " + oData.accept + "\n";
                req += oData.batchChangeSetContentType + "\n\n";

                req += JSON.stringify(arr[i]) + "\n\n";
            }

            return req;
        }
    };

    var queryGen = function (query, adaptor) {
        if (!(this instanceof queryGen))
            return new queryGen(query, adaptor);

        this.query = query;
        this.defaultFilters = "";
        this.adaptor = adaptor;
        return this;
    };

    queryGen.prototype = {
        buildUrl: function (url, requireFormat) {
            var queryStr = "";
            url = this.ensureTable(url);

            queryStr = this.ensureExpands(queryStr);
            queryStr = this.ensureAnd(queryStr);
            queryStr = this.iterateQueries(queryStr);
            queryStr = this.ensureFormat(queryStr, requireFormat);

            if (this.query._requiresCount)
                queryStr += (queryStr.length ? "&" : "") + oData.count + "allpages";

            return url + (queryStr ? "?" + queryStr : "");
        },
        ensureTable: function (url) {
            return url.replace(/\/*$/, this.query._fromTable ? '/' + this.query._fromTable : '');
        },
        ensureExpands: function (url) {
            if (this.query._expands && this.query._expands.length) {
                url += oData.expand;
                for (var j = 0; j < this.query._expands.length; j++) {
                    url += this.query._expands[j];
                    if (j != this.query._expands.length - 1)
                        url += ",";
                }
            }
            return url;
        },
        ensureAnd: function (url, append) {
            if (append !== false && url.length)
                url += "&";
            return url;
        },
        ensureFormat: function (url, reqFormat) {
            if (reqFormat) {
                url = this.ensureAnd(url);
                url += "$format=json";
            }
            return url;
        },
        iterateQueries: function (url) {
            var ret, len = this.query.queries.length, isSorted = false, isFiltered = false, key;

            for (var i = 0; i < len; i++) {
                key = this.query.queries[i];
                ret = this.adaptor[key.fn].call(this, null, this.query, key.e);

                if (!ret) continue;

                if (ret === this.query._sortKey) {
                    if (!isSorted)
                        ret = this.ensureSort();

                    else continue;
                    isSorted = true;
                }

                if (ret === this.query._filterKey) {
                    if (!isFiltered)
                        ret = this.ensureFilter();
                    else continue;

                    isFiltered = true;
                }

                url += ret;
                url = this.ensureAnd(url, i !== len - 1);
            }

            if (!isFiltered && this.defaultFilters) {
                url = this.ensureAnd(url);
                url += this.ensureFilter();
            }
            return url;
        },
        ensureFilter: function () {
            var filter = oData.where, filters = this.query._filters;

            filter += this.defaultFilters;

            if (this.defaultFilters && filters.length)
                filter += " and (";

            for (var i = 0; i < filters.length; i++) {
                if (filters[i] instanceof Array) {
                    var searchQuery = "", tmp;
                    for (var j = 0; j < filters[i].length; j++) {
                        tmp = pvtPredicate.generateODataString(filters[i][j].field, filters[i][j].operator, filters[i][j].value, filters[i][j].ignoreCase);
                        if (!tmp) continue;

                        searchQuery += "(" + tmp + ")";
                        if (j !== filters[i].length - 1)
                            searchQuery += " or ";
                    }
                    filter += searchQuery;
                } else if (filters[i] instanceof sf.Predicate) {
                    filter += filters[i].getQuery();
                } else {
                    filter += pvtPredicate.generateODataString(filters[i].field, filters[i].operator, filters[i].value, filters[i].ignoreCase);
                }

                if (i != filters.length - 1)
                    filter += " and ";
            }

            if (this.defaultFilters && filters.length)
                filter += ")";

            return filter;
        },
        ensureSort: function () {
            var cols = this.query._sortedColumns, query = "";

            for (var i = 0; i < cols.length; i++) {
                query += cols[i];
                if (i !== cols.length - 1)
                    query += ",";
            }

            if (query.length)
                query = oData.sortBy + query;

            return query;
        },
        addFilters: function (data) {
            if (!(data instanceof Array) || !data.length)
                return this;
            var key = this.query._fKey, value, query = "", prop = key;

            if (typeof data[0] !== "object")
                prop = null;

            for (var i = 0; i < data.length; i++) {
                value = sf.pvt.getObject(prop, data[i]);
                query += pvtPredicate.generateODataString(key, "==", value);
                if (i !== data.length - 1)
                    query += " or ";
            }

            this.defaultFilters = "(" + query + ")";
            return this;
        }
    };

    sf.JsonAdaptor = function () {
    };

    sf.JsonAdaptor.prototype = {
        onWhere: function (ds, query, e) {
            if (!ds) return ds;

            return ds.filter(function (obj) {

                if (e.isPredicate)
                    return e.fieldName.validate(obj);

                return e.comparer.call(obj, sf.pvt.getObject(e.fieldName, obj), e.value, e.ignoreCase);
            });
        },
        onSearch: function (ds, query, e) {
            if (!ds || !ds.length) return ds;

            if (e.fieldNames.length === 0) {
                sf.pvt.getFieldList(ds[0], e.fieldNames);
            }

            return ds.filter(function (obj) {
                for (var j = 0; j < e.fieldNames.length; j++) {
                    if (e.comparer.call(obj, sf.pvt.getObject(e.fieldNames[j], obj), e.searchKey, e.ignoreCase))
                        return true;
                }
                return false;
            });
        },
        onSortBy: function (ds, query, e) {
            if (!ds) return ds;
            var fnCompare;
            if (!e.fieldName)
                return ds.sort(e.comparer);

            if (e.isFunc) {
                e.fieldName = e.fieldName.call(query);
            }

            if (e.fieldName instanceof Array) {
                for (var i = e.fieldName.length - 1; i >= 0; i--) {
                    fnCompare = e.comparer;

                    if (e.fieldName[i].endsWith(" desc")) {
                        fnCompare = sf.pvt.fnSort(sf.sortOrder.Descending);
                        e.fieldName[i] = e.fieldName[i].replace(" desc", "");
                    }

                    ds = sf.mergeSort(ds, e.fieldName[i], fnCompare);
                }
                return ds;
            }
            return sf.mergeSort(ds, e.fieldName, e.comparer);
        },
        onGroup: function (ds, query, e) {
            if (!ds) return ds;

            return sf.group(ds, e.fieldName);
        },
        onPage: function (ds, query, e) {
            var start = (e.pageIndex - 1) * e.pageSize, end = start + e.pageSize;

            if (!ds) return ds;

            return ds.slice(start, end);
        },
        onRange: function (ds, query, e) {
            if (!ds) return ds;
            return ds.slice(e.start, e.end);
        },
        onTake: function (ds, query, e) {
            if (!ds) return ds;

            return ds.slice(0, e.nos);
        },
        onSkip: function (ds, query, e) {
            if (!ds) return ds;
            return ds.slice(e.nos);
        },
        onSelect: function (ds, query, e) {
            if (!ds) return ds;
            return sf.select(ds, e.fieldNames);
        }
    };

    sf.TableModel = function (name, jsonArray, dataManager) {
        if (!instance(this, sf.TableModel))
            return new sf.TableModel(jsonArray);

        if (!instance(jsonArray, Array))
            throw "sf.TableModel - Json Array is required";

        var rows = [], model, dirtyFn = $.proxy(setDirty, this);

        for (var i = 0; i < jsonArray.length; i++) {
            model = new sf.Model(jsonArray[i], this);
            model.state = "unchanged";
            model.on("stateChange", dirtyFn);

            rows.push(model);
        }

        this.name = name || "table1";

        this.rows = sf.NotifierArray(rows);
        this._deleted = [];

        this._events = $({});

        this.dataManager = dataManager;

        this._isDirty = false;

        return this;
    };

    sf.TableModel.prototype = {
        on: function (eventName, handler) {
            this._events.on(eventName, handler);
        },

        off: function (eventName, handler) {
            this._events.off(eventName, handler);
        },

        setDataManager: function (dataManager) {
            this.dataManagar = dataManager;
        },

        saveChanges: function () {
            if (!this.dataManager || !instance(this.dataManager, sf.DataManagar))
                throw "sf.TableModel - saveChanges : Set the dataManager using setDataManager function";

            if (!this.isDirty())
                return;

            var promise = this.dataManager.saveChanges(this.getChanges(), this.key, this.name);

            promise.done($.proxy(function () {
                var rows = this.toArray();
                for (var i = 0; i < rows.length; i++) {
                    rows[i].markCommit();
                }
                
                this._events.triggerHandler({ type: "save", table: this });

            }, this));

            promise.fail($.proxy(function (e) {
                this.rejectChanges();
                this._events.triggerHandler({ type: "reject", table: this, error: e });
            }, this));

            this._isDirty = false;
        },

        rejectChanges: function () {
            var rows = this.toArray();
            for (var i = 0; i < rows.length; i++)
                rows[i].revert(true);

            this._isDirty = false;
            this._events.triggerHandler({ type: "reject", table: this });
        },

        insert: function (json) {
            var model = new sf.Model(json);
            model._isDirty = this._isDirty = true;

            this.rows.push(model);

            this._events.triggerHandler({ type: "insert", model: model, table: this });
        },

        update: function (value) {
            if (!this.key)
                throw "TableModel - update : Primary key should be assigned to TableModel.key";

            var row = value, model, key = this.key, keyValue = row[key];

            model = this.rows.array.filter(function (obj) {
                return obj.get(key) === keyValue;
            });

            model = model[0];

            for (var col in row) {
                model.set(col, row[col]);
            }

            this._isDirty = true;

            this._events.triggerHandler({ type: "update", model: model, table: this });
        },

        remove: function (key) {
            if (!this.key)
                throw "TableModel - update : Primary key should be assigned to TableModel.key";

            var field = this.key;

            var index = -1, model;

            if (key && typeof key === "object") {
                key = key[field] !== undefined ? key[field] : key.get(field);
            }

            for (var i = 0; i < this.rows.length(); i++) {
                if (this.rows.array[i].get(field) === key) {
                    index = i;
                    break;
                }
            }

            if (index > -1) {
                model = this.rows.removeAt(index);
                model.markDelete();

                this._deleted.push({ model: model, position: index });

                this._isDirty = true;
                this._events.triggerHandler({ type: "remove", model: model, table: this });
            }
        },

        isDirty: function () {
            return this._isDirty;
        },

        getChanges: function () {

            var changes = {
                added: [],
                changed: []
            };
            var rows = this.toArray();
            for (var i = 0; i < rows.length; i++) {
                if (changes[rows[i].state])
                    changes[rows[i].state].push(rows[i].json);
            }

            changes.deleted = sf.select(this._deleted, ["model"]);

            return changes;
        },

        toArray: function () {
            return this.rows.toArray();
        },

        setDirty: function (dirty, model) {
            if (this._isDirty === !!dirty) return;

            this._isDirty = !!dirty;

            this._events.triggerHandler({ type: "dirty", table: this, model: model });
        }
    };

    sf.Model = function (json, table) {
        this.$id = getUid("m");

        this.json = json;
        this.table = table;
        this.actual = {};
        this._events = $({});
        this.isDirty = false;
        this.state = "added";
        this._props = [];
        this._updates = {};
    };

    sf.Model.prototype = {
        on: function (eventName, handler) {
            this._events.on(eventName, handler);
        },
        off: function (eventName, handler) {
            this._events.off(eventName, handler);
        },
        set: function (field, value) {
            var obj = this.json, actual = field, prev;
            field = field.split('.');

            for (var i = 0; i < field.length - 1; i++) {
                field = field[0];
                obj = obj[field[0]];
            }

            this.isDirty = true;
            this.changeState("changed", { from: "set" });

            prev = obj[field];

            if (this.actual[field] === undefined && !(field in this.actual))
                this.actual[field] = value; // Complex property ?

            obj[field] = value;

            this._updateValues(field, value);
            this._events.triggerHandler({ type: actual, current: value, previous: prev, model: this });
        },
        get: function (field) {
            return sf.pvt.getObject(field, this.json);
        },
        revert: function (suspendEvent) {
            for (var prop in this.actual) {
                this.json[prop] = this.actual[prop];
            }

            this.isDirty = false;

            if (suspendEvent)
                this.state = "unchanged";
            else
                this.changeState("unchanged", { from: "revert" });
        },
        markCommit: function () {
            this.isDirty = false;
            this.changeState("unchanged", { from: "commit" });
        },
        markDelete: function () {
            this.changeState("deleted", { from: "delete" });
        },
        changeState: function (state, args) {
            if (this.state === state) return;

            if (this.state === "added") {
                if (state === "deleted")
                    state = "unchanged";
                else return;
            }

            var prev = state;
            args = args || {};

            this.state = state;
            this._events.triggerHandler($.extend({ type: "stateChange", current: state, previous: prev, model: this }, args));
        },
        properties: function () {
            if (this._props.length)
                return this._props;

            for (var pr in this.json) {
                this._props.push(pr);
                this._updates[pr] = { read: [], input: [] };
            }

            return this._props;
        },
        bindTo: function (element, selector) {
            var props = this.properties(), el = $(element), ctl;

            el.data("sfModel", this);

            for (var j = 0; j < props.length; j++) {

                var input = el.find(selector.replace("{{0}}", props[j])),
                    e = { self: this, prop: props[j] };

                for (var i = 0; i < input.length; i++) {
                    ctl = $(input[i]);

                    if (this.formElements.indexOf(ctl.prop("tagName").toLowerCase()) === -1) {
                        ctl.html(this.get(props[j]));
                        this._updates[props[j]].read.push(ctl);
                        continue;
                    }

                    ctl.val(this.get(props[j]))
                        .off("change", null, this._changeHandler)
                        .on("change", null, e, this._changeHandler);

                    this._updates[props[j]].input.push(ctl);
                }
            }
        },
        unbind: function (element, selector) {
            var props = this.properties(), el = $(element);
            el.removeData("sfModel");

            for (var j = 0; selector && j < props.length; j++) {
                el.find(selector.replace("{{0}}", props[j]))
                    .off("change", null, this._changeHandler);
            }
            this._updateValues = {};
        },
        _updateValues: function (prop, value) {
            var arr = this._updates[prop];

            if (!arr || (!arr.read && !arr.input)) return;

            var ensureItems = function (a, type) {
                for (var i = a.length - 1; i > -1; i--) {
                    if (!a[i][0].offsetParent) {
                        a.splice(i, 1);
                        continue;
                    }
                    a[i][type](value);
                }
            };

            ensureItems(arr.read, "html");
            ensureItems(arr.input, "val");
        },
        _changeHandler: function (e) {
            e.data.self.set(e.data.prop, $(this).val());
        },
        formElements: ["input", "select", "textarea"]
    };

    var setDirty = function (e) {
        this.setDirty(true, e.model);
    };

    sf.Predicate = function (field, operator, value, ignoreCase) {
        if (!(this instanceof sf.Predicate))
            return new sf.Predicate(field, operator, value, ignoreCase);

        this._fn = null;
        this._params = null;

        if (typeof field === "string") {
            this._fn = sf.data.fnOperators.processOperator(operator);

            this._params = {
                field: field,
                value: value,
                operator: operator,
                ignoreCase: ignoreCase
            };
        } else if (field instanceof sf.Predicate && value instanceof sf.Predicate) {
            if (operator && operator == "and" || operator == "or")
                this._fn = sf.data.fnOperators["_" + operator + "predicate"](field, value);
        }
        return this;
    };

    sf.Predicate.and = function () {
        return pvtPredicate._combinePredicates(arguments, "and");
    };

    sf.Predicate.or = function () {
        return pvtPredicate._combinePredicates(arguments, "or");
    };

    // Private fn
    var pvtPredicate = {
        _combinePredicates: function (predicates, operator) {
            if (predicates.length < 2) return undefined;

            var pred = predicates[0];

            for (var i = 1; i < predicates.length; i++) {
                pred = new sf.Predicate(pred, operator, predicates[i]);
            }

            return pred;
        },

        _combine: function (pred, field, operator, value, condition) {
            if (field instanceof sf.Predicate)
                return sf.Predicate[condition](pred, field);

            if (typeof field === "string")
                return sf.Predicate[condition](pred, new sf.Predicate(field, operator, value));

            throw "Predicate - " + condition + " : invalid arguments";
        },

        generateODataString: function (field, operator, value, ignoreCase, requiresCast) {
            var returnValue = "", op = operator, type = typeof value;

            field = toODataProperty(field);
            if (type === "string") {
                value = "'" + value + "'";

                if (requiresCast) {
                    field = "cast(" + field + ", 'Edm.String')";
                }

                if (ignoreCase) {
                    field = "tolower(" + field + ")";
                    value = value.toLowerCase();
                }
            }

            operator = sf.data.odBiOperator[op];
            if (operator) {
                returnValue += field;
                returnValue += operator;

                return returnValue + value;
            }

            operator = sf.data.odUniOperator[op];
            if (!operator || type !== "string") return "";

            if (operator === "substringof") {
                var t = value;
                value = field;
                field = t;
            }

            returnValue += operator + "(";
            returnValue += field + ",";
            returnValue += value + ")";

            return returnValue;
        }
    };

    sf.Predicate.prototype = {
        and: function (field, operator, value) {
            return pvtPredicate._combine(this, field, operator, value, "and");
        },
        or: function (field, operator, value) {
            return pvtPredicate._combine(this, field, operator, value, "or");
        },
        getValidatorFn: function () {
            return this._fn;
        },
        validate: function (record) {

            if (this._params) {
                var value = sf.pvt.getObject(this._params.field, record);
                return this._fn.call(this, value, this._params.value, this._params.ignoreCase);
            }

            return this._fn.validate.call(this, record);
        },
        getQuery: function () {

            if (this._params) {
                return pvtPredicate.generateODataString(this._params.field, this._params.operator, this._params.value, this._params.ignoreCase);
            }

            return this._fn.query.call(this);
        }
    };

    sf.dataUtil = {
        swap: function (array, x, y) {
            if (x == y) return;

            var tmp = array[x];
            array[x] = array[y];
            array[y] = tmp;
        },

        quickSort: function (jsonArray, fieldName, comparer) {

            if (arguments.length == 2 && typeof fieldName === "function") {
                comparer = fieldName;
                fieldName = undefined;
            }

            if (!comparer || typeof comparer === "string") {
                var order = (comparer || "").toLowerCase();
                comparer = order == sf.sortOrder.Descending ? sf.pvt.fnDescending : sf.pvt.fnAscending;
            }

            return sf.pvt.quickSort(jsonArray, fieldName, 0, jsonArray.length - 1, comparer);
        },

        mergeSort: function (jsonArray, fieldName, comparer) {
            if (!comparer || typeof comparer === "string")
                comparer = sf.pvt.fnSort(comparer, true);

            if (typeof fieldName === "function") {
                comparer = fieldName;
                fieldName = null;
            }

            return sf.pvt.mergeSort(jsonArray, fieldName, comparer);
        },

        max: function (jsonArray, fieldName, comparer) {
            if (typeof fieldName === "function") {
                comparer = fieldName;
                fieldName = null;
            }

            return sf.pvt.getItemFromComparer(jsonArray, fieldName, comparer || sf.pvt.fnDescending);
        },

        min: function (jsonArray, fieldName, comparer) {
            if (typeof fieldName === "function") {
                comparer = fieldName;
                fieldName = null;
            }

            return sf.pvt.getItemFromComparer(jsonArray, fieldName, comparer || sf.pvt.fnAscending);
        },

        select: function (jsonArray, fields) {
            var newData = [];

            for (var i = 0; i < jsonArray.length; i++) {
                newData.push(sf.pvt.extractFields(jsonArray[i], fields));
            }

            return newData;
        },

        group: function (jsonArray, field, /* internal */ level) {
            level = level || 1;

            if (jsonArray.GROUPGUID == sf.pvt.consts.GROUPGUID) {
                for (var j = 0; j < jsonArray.length; j++) {
                    jsonArray[j].items = sf.group(jsonArray[j].items, field, jsonArray.level + 1);
                    jsonArray[j].count = jsonArray[j].items.length;
                }

                jsonArray.childLevels += 1;
                return jsonArray;
            }

            var grouped = {}, groupedArray = [];

            groupedArray.GROUPGUID = sf.pvt.consts.GROUPGUID;
            groupedArray.level = level;
            groupedArray.childLevels = 0;

            for (var i = 0; i < jsonArray.length; i++) {
                var val = getVal(jsonArray, field, i);

                if (!grouped[val]) {
                    grouped[val] = {
                        key: val,
                        count: 0,
                        items: [],
                        field: field
                    };
                    groupedArray.push(grouped[val]);
                }

                grouped[val].count += 1;
                grouped[val].items.push(jsonArray[i]);
            }

            return groupedArray;
        }
    };

    sf.pvt = {
        mergeSort: function (jsonArray, fieldName, comparer) {
            if (jsonArray.length <= 1)
                return jsonArray;

            // else list size is > 1, so split the list into two sublists
            var middle = parseInt(jsonArray.length / 2, 10);

            var left = jsonArray.slice(0, middle),
                right = jsonArray.slice(middle);

            left = sf.pvt.mergeSort(left, fieldName, comparer);
            right = sf.pvt.mergeSort(right, fieldName, comparer);

            return sf.pvt.merge(left, right, fieldName, comparer);
        },

        quickSort: function (jsonArray, fieldName, left, right, comparer) {
            if (!comparer) return;

            if (left < right) {

                // partition
                var pivotIndex = sf.pvt.partition(jsonArray, fieldName, left, right, comparer);

                // recursion
                sf.pvt.quickSort(jsonArray, fieldName, left, pivotIndex - 1, comparer);

                sf.pvt.quickSort(jsonArray, fieldName, pivotIndex + 1, right, comparer);
            }
        },

        getItemFromComparer: function (array, field, comparer) {
            var keyVal = getVal(array, field, 0), current, key = array[0];
            for (var i = 1; i < array.length; i++) {
                current = getVal(array, field, i);
                if (comparer(keyVal, current) > 0) {
                    keyVal = current;
                    key = array[i];
                }
            }
            return key;
        },

        quickSelect: function (array, fieldName, left, right, k, comparer) {
            if (left == right)
                return array[left];

            var pivotNewIndex = sf.pvt.partition(array, fieldName, left, right, comparer);

            var pivotDist = pivotNewIndex - left + 1;

            if (pivotDist == k)
                return array[pivotNewIndex];

            else if (k < pivotDist)
                return sf.pvt.quickSelect(array, fieldName, left, pivotNewIndex - 1, k, comparer);
            else
                return sf.pvt.quickSelect(array, fieldName, pivotNewIndex + 1, right, k - pivotDist, comparer);
        },

        extractFields: function (obj, fields) {
            var newObj = {};

            if (fields.length == 1)
                return sf.pvt.getObject(fields[0], obj);

            for (var i = 0; i < fields.length; i++) {
                newObj[fields[i].replace('.', sf.pvt.consts.complexPropertyMerge)] = sf.pvt.getObject(fields[i], obj);
            }

            return newObj;
        },

        partition: function (array, field, left, right, comparer) {

            var pivotIndex = parseInt((left + right) / 2, 10),
                pivot = getVal(array, field, pivotIndex);

            sf.swap(array, pivotIndex, right);

            pivotIndex = left;

            for (var i = left; i < right; i++) {
                if (comparer(getVal(array, field, i), pivot)) { //val(i) < pivot
                    sf.swap(array, i, pivotIndex);
                    pivotIndex++;
                }
            }

            sf.swap(array, pivotIndex, right);

            return pivotIndex;
        },

        fnSort: function (order) {
            order = order ? order.toLowerCase() : sf.sortOrder.Ascending;

            if (order == sf.sortOrder.Ascending)
                return sf.pvt.fnAscending;

            return sf.pvt.fnDescending;
        },

        fnAscending: function (x, y) {
            if (x && x.localeCompare)
                return x.localeCompare(y);
            return x - y;
        },

        fnDescending: function (x, y) {
            if (x && x.localeCompare)
                return x.localeCompare(y) * -1;
            return y - x;
        },

        merge: function (left, right, fieldName, comparer) {
            var result = [], current;

            while (left.length > 0 || right.length > 0) {
                if (left.length > 0 && right.length > 0) {
                    if (comparer)
                        current = comparer(getVal(left, fieldName, 0), getVal(right, fieldName, 0)) < 1 ? left : right;
                    else
                        current = left[0][fieldName] < left[0][fieldName] ? left : right;
                } else {
                    current = left.length > 0 ? left : right;
                }

                result.push(current.shift());
            }

            return result;
        },

        getObject: function (nameSpace, from) {
            if (!from) return undefined;
            if (!nameSpace) return from;

            if (nameSpace.indexOf('.') === -1) return from[nameSpace];

            var value = from, splits = nameSpace.split('.');

            for (var i = 0; i < splits.length; i++) {

                if (value == null) break;

                value = value[splits[i]];
            }

            return value;
        },

        createObject: function (nameSpace, value, initIn) {
            var splits = nameSpace.split('.'), start = initIn || window, from = start, i;

            for (i = 0; i < splits.length; i++) {

                if (i + 1 == splits.length)
                    from[splits[i]] = value === undefined ? {} : value;
                else if (from[splits[i]] == null)
                    from[splits[i]] = {};

                from = from[splits[i]];
            }

            return start;
        },

        getFieldList: function (obj, fields, prefix) {
            if (prefix === undefined)
                prefix = "";

            if (fields === undefined || fields === null)
                return sf.pvt.getFieldList(obj, [], prefix);

            for (var prop in obj) {
                if (typeof obj[prop] === "object" && !(obj[prop] instanceof Array))
                    sf.pvt.getFieldList(obj[prop], fields, prefix + prop + ".");
                else
                    fields.push(prefix + prop);
            }

            return fields;
        }
    };

    sf.FilterOperators = {
        lessThan: "lessthan",
        greaterThan: "greaterthan",
        lessThanOrEqual: "lessthanorequal",
        greaterThanOrEqual: "greaterthanorequal",
        equal: "equal",
        contains: "contains",
        startsWith: "startswith",
        endsWith: "endswith",
        notEqual: "notequal"
    };

    sf.data = {};

    sf.data.operatorSymbols = {
        "<": "lessthan",
        ">": "greaterthan",
        "<=": "lessthanorequal",
        ">=": "greaterthanorequal",
        "==": "equal",
        "!=": "notequal",
        "*=": "contains",
        "$=": "endswith",
        "^=": "startswith"
    };

    sf.data.odBiOperator = {
        "<": " lt ",
        ">": " gt ",
        "<=": " le ",
        ">=": " ge ",
        "==": " eq ",
        "!=": " ne ",
        "lessthan": " lt ",
        "lessthanorequal": " le ",
        "greaterthan": " gt ",
        "greaterthanorequal": " ge ",
        "equal": " eq ",
        "notequal": " ne "
    };

    sf.data.odUniOperator = {
        "$=": "endswith",
        "^=": "startswith",
        "*=": "substringof",
        "endswith": "endswith",
        "startswith": "startswith",
        "contains": "substringof"
    };

    sf.data.fnOperators = {
        equal: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return toLowerCase(actual) == toLowerCase(expected);

            return actual == expected;
        },
        notequal: function (actual, expected, ignoreCase) {
            return !sf.data.fnOperators.equal(actual, expected, ignoreCase);
        },
        lessthan: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return toLowerCase(actual) < toLowerCase(expected);

            return actual < expected;
        },
        greaterthan: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return toLowerCase(actual) > toLowerCase(expected);

            return actual > expected;
        },
        lessthanorequal: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return toLowerCase(actual) <= toLowerCase(expected);

            return actual <= expected;
        },
        greaterthanorequal: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return toLowerCase(actual) >= toLowerCase(expected);

            return actual >= expected;
        },
        contains: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return actual && expected && toLowerCase(actual).indexOf(toLowerCase(expected)) != -1;

            return actual && expected && actual.toString().indexOf(expected) != -1;
        },
        notnull: function (actual) {
            return actual !== null;
        },
        isnull: function (actual) {
            return actual === null;
        },
        startswith: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return actual && expected && toLowerCase(actual).startsWith(toLowerCase(expected));

            return actual && expected && actual.startsWith(expected);
        },
        endswith: function (actual, expected, ignoreCase) {
            if (ignoreCase)
                return actual && expected && toLowerCase(actual).endsWith(toLowerCase(expected));

            return actual && expected && actual.endsWith(expected);
        },

        _andpredicate: function () {
            var args = [].slice.call(arguments, 0);
            return {
                validate: function () {

                    for (var i = 0; i < args.length; i++) {
                        if (!args[i].validate.apply(args[i], arguments))
                            return false;
                    }

                    return true;
                },
                query: function () {
                    var value = "";
                    for (var i = 0; i < args.length; i++) {
                        value += "(" + args[i].getQuery.apply(args[i], arguments) + ")";
                        if (i != args.length - 1)
                            value += " and ";
                    }
                    return value;
                }
            };
        },

        _orpredicate: function () {
            var args = [].slice.call(arguments, 0);
            return {
                validate: function () {

                    for (var i = 0; i < args.length; i++) {
                        if (args[i].validate.apply(args[i], arguments))
                            return true;
                    }

                    return false;
                },
                query: function () {
                    var value = "";
                    for (var i = 0; i < args.length; i++) {
                        value += "(" + args[i].getQuery.apply(args[i], arguments) + ")";
                        if (i != args.length - 1)
                            value += " or ";
                    }
                    return value;
                }
            };
        },

        processSymbols: function (operator) {
            var fnName = sf.data.operatorSymbols[operator];
            if (fnName) {
                var fn = sf.data.fnOperators[fnName];
                if (fn) return fn;
            }

            throw "Query - Process Operator : Invalid operator";
        },

        processOperator: function (operator) {
            var fn = sf.data.fnOperators[operator];
            if (fn) return fn;
            return sf.data.fnOperators.processSymbols(operator);
        }
    };

    sf.NotifierArray = function (array) {
        if (!instance(this, sf.NotifierArray))
            return new sf.NotifierArray(array);

        this.array = array;

        this._events = $({});
        this._isDirty = false;

        return this;
    };

    sf.NotifierArray.prototype = {
        on: function (eventName, handler) {
            this._events.on(eventName, handler);
        },
        off: function (eventName, handler) {
            this._events.off(eventName, handler);
        },
        push: function (item) {
            var ret;

            if (instance(item, Array))
                ret = [].push.apply(this.array, item);
            else
                ret = this.array.push(item);

            this._raise("add", { item: item, index: this.length() - 1 });

            return ret;
        },
        pop: function () {
            var ret = this.array.pop();

            this._raise("remove", { item: ret, index: this.length() - 1 });

            return ret;
        },
        addAt: function (index, item) {
            this.array.splice(index, 0, item);

            this._raise("add", { item: item, index: index });

            return item;
        },
        removeAt: function (index) {
            var ret = this.array.splice(index, 1)[0];

            this._raise("remove", { item: ret, index: index });

            return ret;
        },
        remove: function (item) {
            var index = this.array.indexOf(item);

            if (index > -1) {
                this.array.splice(index, 1);
                this._raise("remove", { item: item, index: index });
            }

            return index;
        },
        length: function () {
            return this.array.length;
        },
        _raise: function (e, args) {
            this._events.triggerHandler($.extend({ type: e }, args));
            this._events.triggerHandler({ type: "all", name: e, args: args });
        },
        toArray: function () {
            return this.array;
        }
    };

    $.extend(sf, sf.dataUtil);

    // For IE8
    Array.prototype.forEach = Array.prototype.forEach || function (fn, scope) {
        for (var i = 0, len = this.length; i < len; ++i) {
            fn.call(scope, this[i], i, this);
        }
    };

    Array.prototype.indexOf = Array.prototype.indexOf || function (searchElement) {
        var len = this.length;

        if (len === 0) return -1;

        for (var i = 0; i < len; i++) {
            if (i in this && this[i] === searchElement)
                return i;
        }
        return -1;
    };

    Array.prototype.filter = Array.prototype.filter || function (fn) {
        return $(this).filter(fn);
    };

    String.prototype.endsWith = String.prototype.endsWith || function (key) {
        return this.slice(-key.length) === key;
    };

    String.prototype.startsWith = String.prototype.startsWith || function (key) {
        return this.slice(0, key.length) === key;
    };

    if (!Object.create) {
        Object.create = function (o) {
            if (arguments.length > 1) {
                throw new Error('Object.create implementation only accepts the first parameter.');
            }

            function f() {
            }

            f.prototype = o;
            return new f();
        };
    }

    sf.sortOrder = {
        Ascending: "ascending",
        Descending: "descending"
    };

    // privates
    sf.pvt.consts = {
        GROUPGUID: "{271bbba0-1ee7}",
        complexPropertyMerge: "_"
    };

    // private utils
    var nextTick = function (fn, context) {
        if (context) fn = $.proxy(fn, context);
        (window.setImmediate || window.setTimeout)(fn, 0);
    };

    var toODataProperty = function (prop) {
        return prop.replace(/\./g, "/");
    };

    var toLowerCase = function (val) {
        return val ? val.toLowerCase ? val.toLowerCase() : val.toString() : val === 0 ? val.toString() : "";
    };

    var getVal = function (array, field, index) {
        return field ? sf.pvt.getObject(field, array[index]) : array[index];
    };

    var isHtmlElement = function (e) {
        return typeof HTMLElement === "object" ? e instanceof HTMLElement :
            e && e.nodeType === 1 && typeof e === "object" && typeof e.nodeName === "string";
    };

    var pushOdSortColumns = function (array, fields, order, isFromGroup) {

        if (typeof fields === "function") {
            array.push(fields);
            return;
        }

        if (typeof fields === "string") {
            order = order === sf.sortOrder.Descending ? " desc" : "";

            if (!fields.endsWith(" desc"))
                fields += order;

            fields = [fields];
        }

        for (var j = 0; j < fields.length; j++) {
            var field = toODataProperty(fields[j]),
                index = array.indexOf(field.replace(" desc", ""));

            if (index !== -1 && isFromGroup)
                array.splice(index, 1);

            if (index === -1 || isFromGroup)
                array[isFromGroup ? "unshift" : "push"](field);
        }
    };

    var instance = function (obj, element) {
        return obj instanceof element;
    };

    var getTableModel = function (name, result, dm) {
        return function () {
            return new sf.TableModel(name, result, dm);
        };
    };

    var uidIndex = 0;
    var getUid = function (prefix) {
        uidIndex += 1;
        return prefix + uidIndex;
    };

    sf.getGuid = function (prefix) {
        var hexs = '0123456789abcdef', rand;
        return prefix + '00000000-0000-4000-0000-000000000000'.replace(/0/g, function (val, i) {
            rand = Math.random() * 16 | 0;
            return hexs[i === 19 ? rand & 0x3 | 0x8 : rand];
        });
    };

    sf.merge = function (first, second) {
        if (!first || !second) return;

        Array.prototype.push.apply(first, second);
    };

})(window.jQuery, window.Syncfusion);